@@ -1,5 +1,5 @@
 /*
-   Copyright (C) 2006-2013 Con Kolivas
+   Copyright (C) 2006-2016,2018,2021 Con Kolivas
    Copyright (C) 2011 Peter Hyman
    Copyright (C) 1998-2003 Andrew Tridgell
 
@@ -47,6 +47,7 @@
 #endif
 #include <math.h>
 #include <utime.h>
+#include <inttypes.h>
 
 #include "md5.h"
 #include "rzip.h"
@@ -104,7 +105,7 @@ i64 get_ram(rzip_control *control)
 	if(!(meminfo = fopen("/proc/meminfo", "r")))
 		fatal_return(("fopen\n"), -1);
 
-	while(!feof(meminfo) && !fscanf(meminfo, "MemTotal: %lld kB", &ramsize)) {
+	while(!feof(meminfo) && !fscanf(meminfo, "MemTotal: %"PRId64" kB", &ramsize)) {
 		if (unlikely(fgets(aux, sizeof(aux), meminfo) == NULL)) {
 			fclose(meminfo);
 			fatal_return(("Failed to fgets in get_ram\n"), -1);
@@ -217,6 +218,9 @@ static bool get_magic(rzip_control *control, char *magic)
 	if ((int) magic[16]) {
 		for (i = 0; i < 5; i++)
 			control->lzma_properties[i] = magic[i + 16];
+		/* Cludge to allow us to read possibly corrupted archives */
+		if (!control->lzma_properties[0])
+			control->lzma_properties[0] = 93;
 	}
 
 	/* Whether this archive contains md5 data at the end or not */
@@ -306,17 +310,14 @@ int open_tmpoutfile(rzip_control *control)
 			fatal_return(("Failed to allocate outfile name\n"), -1);
 		strcpy(control->outfile, control->tmpdir);
 		strcat(control->outfile, "lrzipout.XXXXXX");
-	} else {
-		control->outfile = realloc(NULL, 16);
-		if (unlikely(!control->outfile))
-			fatal_return(("Failed to allocate outfile name\n"), -1);
-		strcpy(control->outfile, "lrzipout.XXXXXX");
 	}
 
 	fd_out = mkstemp(control->outfile);
-	if (unlikely(fd_out == -1))
-		fatal_return(("Failed to create out tmpfile: %s\n", control->outfile), -1);
-	register_outfile(control, control->outfile, TEST_ONLY || STDOUT || !KEEP_BROKEN);
+	if (fd_out == -1) {
+		print_progress("WARNING: Failed to create out tmpfile: %s, will fail if cannot perform %scompression entirely in ram\n",
+			       control->outfile, DECOMPRESS ? "de" : "");
+	} else
+		register_outfile(control, control->outfile, TEST_ONLY || STDOUT || !KEEP_BROKEN);
 	return fd_out;
 }
 
@@ -328,12 +329,14 @@ static bool fwrite_stdout(rzip_control *control, void *buf, i64 len)
 
 	total = 0;
 	while (len > 0) {
+		ssize_t wrote;
+
 		if (len > one_g)
 			ret = one_g;
 		else
 			ret = len;
-		ret = fwrite(offset_buf, 1, ret, control->outFILE);
-		if (unlikely(ret <= 0))
+		wrote = fwrite(offset_buf, 1, ret, control->outFILE);
+		if (unlikely(wrote != ret))
 			fatal_return(("Failed to fwrite in fwrite_stdout\n"), false);
 		len -= ret;
 		offset_buf += ret;
@@ -352,7 +355,7 @@ bool write_fdout(rzip_control *control, void *buf, i64 len)
 		ret = MIN(len, one_g);
 		ret = write(control->fd_out, offset_buf, (size_t)ret);
 		if (unlikely(ret <= 0))
-			fatal_return(("Failed to write to fd_out in write_fdout\n"), false);;
+			fatal_return(("Failed to write to fd_out in write_fdout\n"), false);
 		len -= ret;
 		offset_buf += ret;
 	}
@@ -382,6 +385,8 @@ bool dump_tmpoutfile(rzip_control *control, int fd_out)
 	FILE *tmpoutfp;
 	int tmpchar;
 
+	if (unlikely(fd_out == -1))
+		fatal_return(("Failed: No temporary outfile created, unable to do in ram\n"), false);
 	/* flush anything not yet in the temporary file */
 	fsync(fd_out);
 	tmpoutfp = fdopen(fd_out, "r");
@@ -424,31 +429,50 @@ bool write_fdin(rzip_control *control)
 /* Open a temporary inputfile to perform stdin decompression */
 int open_tmpinfile(rzip_control *control)
 {
-	int fd_in;
+	int fd_in = -1;
 
+	/* Use temporary directory if there is one */
 	if (control->tmpdir) {
 		control->infile = malloc(strlen(control->tmpdir) + 15);
 		if (unlikely(!control->infile))
 			fatal_return(("Failed to allocate infile name\n"), -1);
 		strcpy(control->infile, control->tmpdir);
 		strcat(control->infile, "lrzipin.XXXXXX");
-	} else {
-		control->infile = malloc(15);
+		fd_in = mkstemp(control->infile);
+	}
+
+	/* Try the current directory */
+	if (fd_in == -1) {
+		dealloc(control->infile);
+		control->infile = malloc(16);
 		if (unlikely(!control->infile))
 			fatal_return(("Failed to allocate infile name\n"), -1);
 		strcpy(control->infile, "lrzipin.XXXXXX");
+		fd_in = mkstemp(control->infile);
 	}
 
-	fd_in = mkstemp(control->infile);
-	if (unlikely(fd_in == -1))
-		fatal_return(("Failed to create in tmpfile: %s\n", control->infile), -1);
-	register_infile(control, control->infile, (DECOMPRESS || TEST_ONLY) && STDIN);
-	/* Unlink temporary file immediately to minimise chance of files left
-	 * lying around in cases of failure_return((. */
-	if (unlikely(unlink(control->infile))) {
-		fatal("Failed to unlink tmpfile: %s\n", control->infile);
-		close(fd_in);
-		return -1;
+	/* Use /tmp if nothing is writeable so far */
+	if (fd_in == -1) {
+		dealloc(control->infile);
+		control->infile = malloc(20);
+		if (unlikely(!control->infile))
+			fatal_return(("Failed to allocate infile name\n"), -1);
+		strcpy(control->infile, "/tmp/lrzipin.XXXXXX");
+		fd_in = mkstemp(control->infile);
+	}
+
+	if (fd_in == -1) {
+		print_progress("WARNING: Failed to create in tmpfile: %s, will fail if cannot perform %scompression entirely in ram\n",
+			       control->infile, DECOMPRESS ? "de" : "");
+	} else {
+		register_infile(control, control->infile, (DECOMPRESS || TEST_ONLY) && STDIN);
+		/* Unlink temporary file immediately to minimise chance of files left
+		* lying around in cases of failure_return((. */
+		if (unlikely(unlink(control->infile))) {
+			fatal("Failed to unlink tmpfile: %s\n", control->infile);
+			close(fd_in);
+			return -1;
+		}
 	}
 	return fd_in;
 }
@@ -474,6 +498,8 @@ bool read_tmpinfile(rzip_control *control, int fd_in)
 	FILE *tmpinfp;
 	int tmpchar;
 
+	if (fd_in == -1)
+		return false;
 	if (control->flags & FLAG_SHOW_PROGRESS)
 		fprintf(control->msgout, "Copying from stdin.\n");
 	tmpinfp = fdopen(fd_in, "w+");
@@ -492,22 +518,39 @@ bool read_tmpinfile(rzip_control *control, int fd_in)
  * a pseudo-temporary file */
 static bool open_tmpoutbuf(rzip_control *control)
 {
+	i64 maxlen = control->maxram;
+	void *buf;
+
+	while (42) {
+		round_to_page(&maxlen);
+		buf = malloc(maxlen);
+		if (buf) {
+			print_maxverbose("Malloced %"PRId64" for tmp_outbuf\n", maxlen);
+			break;
+		}
+		maxlen = maxlen / 3 * 2;
+		if (maxlen < 100000000)
+			fatal_return(("Unable to even malloc 100MB for tmp_outbuf\n"), false);
+	}
 	control->flags |= FLAG_TMP_OUTBUF;
-	control->out_maxlen = control->maxram;
 	/* Allocate slightly more so we can cope when the buffer overflows and
 	 * fall back to a real temporary file */
-	control->tmp_outbuf = malloc(control->maxram + control->page_size);
-	if (unlikely(!control->tmp_outbuf))
-		fatal_return(("Failed to malloc tmp_outbuf in open_tmpoutbuf\n"), false);
+	control->out_maxlen = maxlen - control->page_size;
+	control->tmp_outbuf = buf;
 	if (!DECOMPRESS && !TEST_ONLY)
 		control->out_ofs = control->out_len = MAGIC_LEN;\
 	return true;
 }
 
+/* We've decided to use a temporary output file instead of trying to store
+ * all the output buffer in ram so we can free up the ram and increase the
+ * maximum sizes of ram we can allocate */
 void close_tmpoutbuf(rzip_control *control)
 {
 	control->flags &= ~FLAG_TMP_OUTBUF;
-	free(control->tmp_outbuf);
+	dealloc(control->tmp_outbuf);
+	if (!BITS32)
+		control->usable_ram = control->maxram += control->ramsize / 18;
 }
 
 static bool open_tmpinbuf(rzip_control *control)
@@ -534,10 +577,13 @@ bool clear_tmpinfile(rzip_control *control)
 	return true;
 }
 
+/* As per temporary output file but for input file */
 void close_tmpinbuf(rzip_control *control)
 {
 	control->flags &= ~FLAG_TMP_INBUF;
-	free(control->tmp_inbuf);
+	dealloc(control->tmp_inbuf);
+	if (!BITS32)
+		control->usable_ram = control->maxram += control->ramsize / 18;
 }
 
 static int get_pass(rzip_control *control, char *s)
@@ -545,7 +591,9 @@ static int get_pass(rzip_control *control, char *s)
 	int len;
 
 	memset(s, 0, PASS_LEN - SALT_LEN);
-	if (unlikely(fgets(s, PASS_LEN - SALT_LEN, stdin) == NULL))
+	if (control->passphrase)
+		strncpy(s, control->passphrase, PASS_LEN - SALT_LEN - 1);
+	else if (unlikely(fgets(s, PASS_LEN - SALT_LEN, stdin) == NULL))
 		failure_return(("Failed to retrieve passphrase\n"), -1);
 	len = strlen(s);
 	if (len > 0 && ('\r' ==  s[len - 1] || '\n' == s[len - 1]))
@@ -562,6 +610,7 @@ static bool get_hash(rzip_control *control, int make_hash)
 {
 	char *passphrase, *testphrase;
 	struct termios termios_p;
+	int prompt = control->passphrase == NULL;
 
 	passphrase = calloc(PASS_LEN, 1);
 	testphrase = calloc(PASS_LEN, 1);
@@ -569,8 +618,8 @@ static bool get_hash(rzip_control *control, int make_hash)
 	control->hash = calloc(HASH_LEN, 1);
 	if (unlikely(!passphrase || !testphrase || !control->salt_pass || !control->hash)) {
 		fatal("Failed to calloc encrypt buffers in compress_file\n");
-		free(testphrase);
-		free(passphrase);
+		dealloc(testphrase);
+		dealloc(passphrase);
 		return false;
 	}
 	mlock(passphrase, PASS_LEN);
@@ -584,8 +633,8 @@ static bool get_hash(rzip_control *control, int make_hash)
 			fatal("Supplied password was null!");
 			munlock(passphrase, PASS_LEN);
 			munlock(testphrase, PASS_LEN);
-			free(testphrase);
-			free(passphrase);
+			dealloc(testphrase);
+			dealloc(passphrase);
 			release_hashes(control);
 			return false;
 		}
@@ -596,13 +645,17 @@ static bool get_hash(rzip_control *control, int make_hash)
 		termios_p.c_lflag &= ~ECHO;
 		tcsetattr(fileno(stdin), 0, &termios_p);
 retry_pass:
-		print_output("Enter passphrase: ");
+		if (prompt)
+			print_output("Enter passphrase: ");
 		control->salt_pass_len = get_pass(control, passphrase) + SALT_LEN;
-		print_output("\n");
-		if (make_hash) {
-			print_output("Re-enter passphrase: ");
-			get_pass(control, testphrase);
+		if (prompt)
 			print_output("\n");
+		if (make_hash) {
+			if (prompt)
+				print_output("Re-enter passphrase: ");
+			get_pass(control, testphrase);
+			if (prompt)
+				print_output("\n");
 			if (strcmp(passphrase, testphrase)) {
 				print_output("Passwords do not match. Try again.\n");
 				goto retry_pass;
@@ -618,8 +671,8 @@ retry_pass:
 	memset(passphrase, 0, PASS_LEN);
 	munlock(passphrase, PASS_LEN);
 	munlock(testphrase, PASS_LEN);
-	free(testphrase);
-	free(passphrase);
+	dealloc(testphrase);
+	dealloc(passphrase);
 	return true;
 }
 
@@ -629,8 +682,8 @@ static void release_hashes(rzip_control *control)
 	memset(control->hash, 0, SALT_LEN);
 	munlock(control->salt_pass, PASS_LEN);
 	munlock(control->hash, HASH_LEN);
-	free(control->salt_pass);
-	free(control->hash);
+	dealloc(control->salt_pass);
+	dealloc(control->hash);
 }
 
 /*
@@ -643,7 +696,7 @@ bool decompress_file(rzip_control *control)
 	i64 expected_size = 0, free_space;
 	struct statvfs fbuf;
 
-	if (!STDIN) {
+	if (!STDIN && !IS_FROM_FILE) {
 		struct stat fdin_stat;
 
 		stat(control->infile, &fdin_stat);
@@ -693,10 +746,11 @@ bool decompress_file(rzip_control *control)
 			print_progress("Output filename is: %s\n", control->outfile);
 	}
 
-	if (STDIN) {
+	if ( IS_FROM_FILE ) {
+		fd_in = fileno(control->inFILE);
+	}
+	else if (STDIN) {
 		fd_in = open_tmpinfile(control);
-		if (unlikely(fd_in == -1))
-			return false;
 		read_tmpinmagic(control);
 		if (ENCRYPT)
 			failure_return(("Cannot decompress encrypted file from STDIN\n"), false);
@@ -734,24 +788,31 @@ bool decompress_file(rzip_control *control)
 				return false;
 	} else {
 		fd_out = open_tmpoutfile(control);
-		if (unlikely(fd_out == -1))
-			fatal_return(("Failed to create %s\n", control->outfile), false);
-		fd_hist = open(control->outfile, O_RDONLY);
-		if (unlikely(fd_hist == -1))
-			fatal_return(("Failed to open history file %s\n", control->outfile), false);
-		/* Unlink temporary file as soon as possible */
-		if (unlikely(unlink(control->outfile)))
-			fatal_return(("Failed to unlink tmpfile: %s\n", control->outfile), false);
+		if (fd_out == -1) {
+			fd_hist = -1;
+		} else {
+			fd_hist = open(control->outfile, O_RDONLY);
+			if (unlikely(fd_hist == -1))
+				fatal_return(("Failed to open history file %s\n", control->outfile), false);
+			/* Unlink temporary file as soon as possible */
+			if (unlikely(unlink(control->outfile)))
+				fatal_return(("Failed to unlink tmpfile: %s\n", control->outfile), false);
+		}
 	}
 
-	if (unlikely(!open_tmpoutbuf(control)))
-		return false;
+	if (STDOUT) {
+		if (unlikely(!open_tmpoutbuf(control)))
+			return false;
+	}
 
-	if (!STDIN)
+	if (!STDIN) {
 		if (unlikely(!read_magic(control, fd_in, &expected_size)))
 			return false;
+		if (unlikely(expected_size < 0))
+			fatal_return(("Invalid expected size %lld\n", expected_size), false);
+	}
 
-	if (!STDOUT) {
+	if (!STDOUT && !TEST_ONLY) {
 		/* Check if there's enough free space on the device chosen to fit the
 		* decompressed file. */
 		if (unlikely(fstatvfs(fd_out, &fbuf)))
@@ -784,14 +845,15 @@ bool decompress_file(rzip_control *control)
 	if (unlikely(runzip_fd(control, fd_in, fd_out, fd_hist, expected_size) < 0))
 		return false;
 
-	if (STDOUT && !TMP_OUTBUF)
+	if (STDOUT && !TMP_OUTBUF) {
 		if (unlikely(!dump_tmpoutfile(control, fd_out)))
 			return false;
+	}
 
 	/* if we get here, no fatal_return(( errors during decompression */
 	print_progress("\r");
 	if (!(STDOUT | TEST_ONLY))
-		print_output("Output filename is: %s: ", control->outfile);
+		print_progress("Output filename is: %s: ", control->outfile);
 	if (!expected_size)
 		expected_size = control->st_size;
 	if (!ENCRYPT)
@@ -799,15 +861,22 @@ bool decompress_file(rzip_control *control)
 	else
 		print_progress("[OK]                                             \n");
 
-	if (unlikely(close(fd_hist) || close(fd_out)))
-		fatal_return(("Failed to close files\n"), false);
+	if (TMP_OUTBUF)
+		close_tmpoutbuf(control);
+
+	if (fd_out > 0) {
+		if (unlikely(close(fd_hist) || close(fd_out)))
+			fatal_return(("Failed to close files\n"), false);
+	}
 
 	if (unlikely(!STDIN && !STDOUT && !TEST_ONLY && !preserve_times(control, fd_in)))
 		return false;
 
-	close(fd_in);
+	if ( ! IS_FROM_FILE ) {
+		close(fd_in);
+	}
 
-	if (!KEEP_FILES) {
+	if (!KEEP_FILES && !STDIN) {
 		if (unlikely(unlink(control->infile)))
 			fatal_return(("Failed to unlink %s\n", infilecopy), false);
 	}
@@ -815,7 +884,7 @@ bool decompress_file(rzip_control *control)
 	if (ENCRYPT)
 		release_hashes(control);
 
-	free(control->outfile);
+	dealloc(control->outfile);
 	return true;
 }
 
@@ -867,7 +936,9 @@ static double percentage(i64 num, i64 den)
 
 	if (den < 100) {
 		d_num = num * 100;
-		d_den = den ? : 1;
+		d_den = den;
+		if (!d_den)
+			d_den = 1;
 	} else {
 		d_num = num;
 		d_den = den / 100;
@@ -877,13 +948,14 @@ static double percentage(i64 num, i64 den)
 
 bool get_fileinfo(rzip_control *control)
 {
-	i64 u_len, c_len, last_head, utotal = 0, ctotal = 0, ofs = 25, stream_head[2];
+	i64 u_len, c_len, second_last, last_head, utotal = 0, ctotal = 0, ofs = 25, stream_head[2];
 	i64 expected_size, infile_size, chunk_size = 0, chunk_total = 0;
 	int header_length, stream = 0, chunk = 0;
 	char *tmp, *infilecopy = NULL;
 	char chunk_byte = 0;
 	long double cratio;
 	uchar ctype = 0;
+	uchar save_ctype = 255;
 	struct stat st;
 	int fd_in;
 
@@ -900,7 +972,9 @@ bool get_fileinfo(rzip_control *control)
 			infilecopy = strdupa(control->infile);
 	}
 
-	if (STDIN)
+	if ( IS_FROM_FILE )
+		fd_in = fileno(control->inFILE);
+	else if (STDIN)
 		fd_in = 0;
 	else {
 		fd_in = open(infilecopy, O_RDONLY);
@@ -914,23 +988,29 @@ bool get_fileinfo(rzip_control *control)
 	infile_size = st.st_size;
 
 	/* Get decompressed size */
-	if (unlikely(!read_magic(control, fd_in, &expected_size))) goto error;
+	if (unlikely(!read_magic(control, fd_in, &expected_size)))
+		goto error;
 
 	if (ENCRYPT) {
 		print_output("Encrypted lrzip archive. No further information available\n");
-		if (!STDIN) close(fd_in);
+		if (!STDIN && !IS_FROM_FILE)
+			close(fd_in);
 		goto out;
 	}
 
 	if (control->major_version == 0 && control->minor_version > 4) {
 		if (unlikely(read(fd_in, &chunk_byte, 1) != 1))
 			fatal_goto(("Failed to read chunk_byte in get_fileinfo\n"), error);
+		if (unlikely(chunk_byte < 1 || chunk_byte > 8))
+			fatal_goto(("Invalid chunk bytes %d\n", chunk_byte), error);
 		if (control->major_version == 0 && control->minor_version > 5) {
 			if (unlikely(read(fd_in, &control->eof, 1) != 1))
 				fatal_goto(("Failed to read eof in get_fileinfo\n"), error);
 			if (unlikely(read(fd_in, &chunk_size, chunk_byte) != chunk_byte))
 				fatal_goto(("Failed to read chunk_size in get_fileinfo\n"), error);
 			chunk_size = le64toh(chunk_size);
+			if (unlikely(chunk_size < 0))
+				fatal_goto(("Invalid chunk size %lld\n", chunk_size), error);
 		}
 	}
 
@@ -962,10 +1042,13 @@ next_chunk:
 		chunk_total += chunk_size;
 		print_verbose("Chunk size: %lld\n", chunk_size);
 	}
+	if (unlikely(chunk_byte && (chunk_byte > 8 || chunk_size < 0)))
+		failure("Invalid chunk data\n");
 	while (stream < NUM_STREAMS) {
 		int block = 1;
 
-		if (unlikely(lseek(fd_in, stream_head[stream] + ofs, SEEK_SET)) == -1)
+		second_last = 0;
+		if (unlikely(lseek(fd_in, stream_head[stream] + ofs, SEEK_SET) == -1))
 			fatal_goto(("Failed to seek to header data in get_fileinfo\n"), error);
 		if (unlikely(!get_header_info(control, fd_in, &ctype, &c_len, &u_len, &last_head, chunk_byte)))
 			return false;
@@ -976,9 +1059,12 @@ next_chunk:
 		do {
 			i64 head_off;
 
+			if (unlikely(last_head && last_head < second_last))
+				failure_goto(("Invalid earlier last_head position, corrupt archive.\n"), error);
+			second_last = last_head;
 			if (unlikely(last_head + ofs > infile_size))
 				failure_goto(("Offset greater than archive size, likely corrupted/truncated archive.\n"), error);
-			if (unlikely(head_off = lseek(fd_in, last_head + ofs, SEEK_SET)) == -1)
+			if (unlikely((head_off = lseek(fd_in, last_head + ofs, SEEK_SET)) == -1))
 				fatal_goto(("Failed to seek to header data in get_fileinfo\n"), error);
 			if (unlikely(!get_header_info(control, fd_in, &ctype, &c_len, &u_len,
 					&last_head, chunk_byte)))
@@ -1000,6 +1086,11 @@ next_chunk:
 				print_verbose("zpaq");
 			else
 				print_verbose("Dunno wtf");
+			if (save_ctype == 255)
+				save_ctype = ctype; /* need this for lzma when some chunks could have no compression
+						     * and info will show rzip + none on info display if last chunk
+						     * is not compressed. Adjust for all types in case it's used in
+						     * the future */
 			utotal += u_len;
 			ctotal += c_len;
 			print_verbose("\t%.1f%%\t%lld / %lld", percentage(c_len, u_len), c_len, u_len);
@@ -1019,6 +1110,8 @@ next_chunk:
 	if (control->major_version == 0 && control->minor_version > 4) {
 		if (unlikely(read(fd_in, &chunk_byte, 1) != 1))
 			fatal_goto(("Failed to read chunk_byte in get_fileinfo\n"), error);
+		if (unlikely(chunk_byte < 1 || chunk_byte > 8))
+			fatal_goto(("Invalid chunk bytes %d\n", chunk_byte), error);
 		ofs++;
 		if (control->major_version == 0 && control->minor_version > 5) {
 			if (unlikely(read(fd_in, &control->eof, 1) != 1))
@@ -1026,6 +1119,8 @@ next_chunk:
 			if (unlikely(read(fd_in, &chunk_size, chunk_byte) != chunk_byte))
 				fatal_goto(("Failed to read chunk_size in get_fileinfo\n"), error);
 			chunk_size = le64toh(chunk_size);
+			if (unlikely(chunk_size < 0))
+				fatal_goto(("Invalid chunk size %lld\n", chunk_size), error);
 			ofs += 1 + chunk_byte;
 			header_length = 1 + (chunk_byte * 3);
 		}
@@ -1034,8 +1129,6 @@ next_chunk:
 done:
 	if (unlikely(ofs > infile_size))
 		failure_goto(("Offset greater than archive size, likely corrupted/truncated archive.\n"), error);
-	if (chunk_total > expected_size)
-		expected_size = chunk_total;
 	print_verbose("Rzip compression: %.1f%% %lld / %lld\n",
 			percentage (utotal, expected_size),
 			utotal, expected_size);
@@ -1051,17 +1144,17 @@ done:
 	print_output("%s:\nlrzip version: %d.%d file\n", infilecopy, control->major_version, control->minor_version);
 
 	print_output("Compression: ");
-	if (ctype == CTYPE_NONE)
+	if (save_ctype == CTYPE_NONE)
 		print_output("rzip alone\n");
-	else if (ctype == CTYPE_BZIP2)
+	else if (save_ctype == CTYPE_BZIP2)
 		print_output("rzip + bzip2\n");
-	else if (ctype == CTYPE_LZO)
+	else if (save_ctype == CTYPE_LZO)
 		print_output("rzip + lzo\n");
-	else if (ctype == CTYPE_LZMA)
+	else if (save_ctype == CTYPE_LZMA)
 		print_output("rzip + lzma\n");
-	else if (ctype == CTYPE_GZIP)
+	else if (save_ctype == CTYPE_GZIP)
 		print_output("rzip + gzip\n");
-	else if (ctype == CTYPE_ZPAQ)
+	else if (save_ctype == CTYPE_ZPAQ)
 		print_output("rzip + zpaq\n");
 	else
 		print_output("Dunno wtf\n");
@@ -1074,7 +1167,7 @@ done:
 		int i;
 
 		print_output("MD5 used for integrity testing\n");
-		if (unlikely(lseek(fd_in, -MD5_DIGEST_SIZE, SEEK_END)) == -1)
+		if (unlikely(lseek(fd_in, -MD5_DIGEST_SIZE, SEEK_END) == -1))
 			fatal_goto(("Failed to seek to md5 data in runzip_fd\n"), error);
 		if (unlikely(read(fd_in, md5_stored, MD5_DIGEST_SIZE) != MD5_DIGEST_SIZE))
 			fatal_goto(("Failed to read md5 data in runzip_fd\n"), error);
@@ -1084,14 +1177,15 @@ done:
 		print_output("\n");
 	} else
 		print_output("CRC32 used for integrity testing\n");
-	if (unlikely(close(fd_in)))
-		fatal_return(("Failed to close fd_in in get_fileinfo\n"), false);
+	if ( !IS_FROM_FILE )
+		if (unlikely(close(fd_in)))
+			fatal_return(("Failed to close fd_in in get_fileinfo\n"), false);
 
 out:
-	free(control->outfile);
+	dealloc(control->outfile);
 	return true;
 error:
-	if (!STDIN) close(fd_in);
+	if (!STDIN && ! IS_FROM_FILE) close(fd_in);
 	return false;
 }
 
@@ -1113,17 +1207,20 @@ bool compress_file(rzip_control *control)
 			return false;
 	memset(header, 0, sizeof(header));
 
-	if (!STDIN) {
-		/* is extension at end of infile? */
+	if ( IS_FROM_FILE )
+		fd_in = fileno(control->inFILE);
+	else if (!STDIN) {
+		 /* is extension at end of infile? */
 		if ((tmp = strrchr(control->infile, '.')) && !strcmp(tmp, control->suffix)) {
 			print_err("%s: already has %s suffix. Skipping...\n", control->infile, control->suffix);
 			return false;
 		}
 
-		fd_in = open(control->infile, O_RDONLY);
+        fd_in = open(control->infile, O_RDONLY);
 		if (unlikely(fd_in == -1))
 			fatal_return(("Failed to open %s\n", control->infile), false);
-	} else
+	} 
+	else
 		fd_in = 0;
 
 	if (!STDOUT) {
@@ -1176,10 +1273,14 @@ bool compress_file(rzip_control *control)
 			fatal_goto(("Failed to create %s\n", control->outfile), error);
 		}
 		control->fd_out = fd_out;
-		if (!STDIN)
-			if (unlikely(!preserve_perms(control, fd_in, fd_out))) goto error;
-	} else
-		if (unlikely(!open_tmpoutbuf(control))) goto error;
+		if (!STDIN) {
+			if (unlikely(!preserve_perms(control, fd_in, fd_out)))
+				goto error;
+		}
+	} else {
+		if (unlikely(!open_tmpoutbuf(control)))
+			goto error;
+	}
 
 	/* Write zeroes to header at beginning of file */
 	if (unlikely(!STDOUT && write(fd_out, header, sizeof(header)) != sizeof(header)))
@@ -1188,12 +1289,19 @@ bool compress_file(rzip_control *control)
 	rzip_fd(control, fd_in, fd_out);
 
 	/* Wwrite magic at end b/c lzma does not tell us properties until it is done */
-	if (!STDOUT)
-		if (unlikely(!write_magic(control))) goto error;
+	if (!STDOUT) {
+		if (unlikely(!write_magic(control)))
+			goto error;
+	}
 
 	if (ENCRYPT)
 		release_hashes(control);
 
+	if (unlikely(!STDIN && !STDOUT && !preserve_times(control, fd_in))) {
+		fatal("Failed to preserve times on output file\n");
+		goto error;
+	}
+
 	if (unlikely(close(fd_in))) {
 		fatal("Failed to close fd_in\n");
 		fd_in = -1;
@@ -1204,30 +1312,33 @@ bool compress_file(rzip_control *control)
 	if (TMP_OUTBUF)
 		close_tmpoutbuf(control);
 
-	if (!KEEP_FILES) {
+	if (!KEEP_FILES && !STDIN) {
 		if (unlikely(unlink(control->infile)))
 			fatal_return(("Failed to unlink %s\n", control->infile), false);
 	}
 
-	free(control->outfile);
+	dealloc(control->outfile);
 	return true;
 error:
-	if (STDIN && (fd_in > 0)) close(fd_in);
-	if ((!STDOUT) && (fd_out > 0)) close(fd_out);
+	if (! IS_FROM_FILE && STDIN && (fd_in > 0))
+		close(fd_in);
+	if ((!STDOUT) && (fd_out > 0))
+		close(fd_out);
 	return false;
 }
 
-bool initialize_control(rzip_control *control)
+bool initialise_control(rzip_control *control)
 {
-	struct timeval tv;
-	char *eptr; /* for environment */
+	time_t now_t, tdiff;
+	char localeptr[] = "./", *eptr; /* for environment */
+	size_t len;
 
 	memset(control, 0, sizeof(rzip_control));
 	control->msgout = stderr;
 	control->msgerr = stderr;
 	register_outputfile(control, control->msgout);
 	control->flags = FLAG_SHOW_PROGRESS | FLAG_KEEP_FILES | FLAG_THRESHOLD;
-	control->suffix = strdup(".lrz");
+	control->suffix = ".lrz";
 	control->compression_level = 7;
 	control->ramsize = get_ram(control);
 	if (unlikely(control->ramsize == -1))
@@ -1240,24 +1351,40 @@ bool initialize_control(rzip_control *control)
 	/* The first 5 bytes of the salt is the time in seconds.
 	 * The next 2 bytes encode how many times to hash the password.
 	 * The last 9 bytes are random data, making 16 bytes of salt */
-	if (unlikely(gettimeofday(&tv, NULL)))
-		fatal_return(("Failed to gettimeofday in main\n"), false);
-	control->secs = tv.tv_sec;
+	if (unlikely((now_t = time(NULL)) == ((time_t)-1)))
+		fatal_return(("Failed to call time in main\n"), false);
+	if (unlikely(now_t < T_ZERO)) {
+		print_output("Warning your time reads before the year 2011, check your system clock\n");
+		now_t = T_ZERO;
+	}
+	/* Workaround for CPUs no longer keeping up with Moore's law!
+	 * This way we keep the magic header format unchanged. */
+	tdiff = (now_t - T_ZERO) / 4;
+	now_t = T_ZERO + tdiff;
+	control->secs = now_t;
 	control->encloops = nloops(control->secs, control->salt, control->salt + 1);
 	if (unlikely(!get_rand(control, control->salt + 2, 6)))
 		return false;
 
-	/* Get Temp Dir */
-	eptr = getenv("TMP");
-	if (eptr != NULL) {
-		size_t len = strlen(eptr);
-		control->tmpdir = malloc(len+2);
-		if (control->tmpdir == NULL)
-			fatal_return(("Failed to allocate for tmpdir\n"), false);
-		strcpy(control->tmpdir, eptr);
-		if (eptr[len - 2] != '/')
-			eptr[len - 2] = '/'; /* need a trailing slash */
-		eptr[len - 1] = 0;
+	/* Get Temp Dir. Try variations on canonical unix environment variable */
+	eptr = getenv("TMPDIR");
+	if (!eptr)
+		eptr = getenv("TMP");
+	if (!eptr)
+		eptr = getenv("TEMPDIR");
+	if (!eptr)
+		eptr = getenv("TEMP");
+	if (!eptr)
+		eptr = localeptr;
+	len = strlen(eptr);
+
+	control->tmpdir = malloc(len + 2);
+	if (control->tmpdir == NULL)
+		fatal_return(("Failed to allocate for tmpdir\n"), false);
+	strcpy(control->tmpdir, eptr);
+	if (control->tmpdir[len - 1] != '/') {
+		control->tmpdir[len] = '/'; /* need a trailing slash */
+		control->tmpdir[len + 1] = '\0';
 	}
 	return true;
 }
