@@ -1,5 +1,5 @@
 /*
-   Copyright (C) 2006-2015 Con Kolivas
+   Copyright (C) 2006-2016,2018,2021 Con Kolivas
    Copyright (C) 2011 Peter Hyman
    Copyright (C) 1998-2003 Andrew Tridgell
 
@@ -67,6 +67,10 @@ void *alloca (size_t);
 # include <sys/endian.h>
 #endif
 #ifndef __BYTE_ORDER
+# ifndef __BIG_ENDIAN
+#  define __BIG_ENDIAN	4321
+#  define __LITTLE_ENDIAN	1234
+# endif
 # ifdef WORDS_BIGENDIAN
 #  define __BYTE_ORDER __BIG_ENDIAN
 # else
@@ -140,7 +144,7 @@ extern int errno;
 #define unlikely(x)	__builtin_expect(!!(x), 0)
 #define __maybe_unused	__attribute__((unused))
 
-#if defined(__MINGW32__) || defined(__CYGWIN__) || defined(ANDROID) || defined(__APPLE__)
+#if defined(__MINGW32__) || defined(__CYGWIN__) || defined(__ANDROID__) || defined(__APPLE__)
 # define ffsll __builtin_ffsll
 #endif
 
@@ -262,6 +266,11 @@ typedef sem_t cksem_t;
 # define PAGE_SIZE (4096)
 #endif
 
+#define dealloc(ptr) do { \
+	free(ptr); \
+	ptr = NULL; \
+} while (0)
+
 /* Determine how many times to hash the password when encrypting, based on
  * the date such that we increase the number of loops according to Moore's
  * law relative to when the data is encrypted. It is then stored as a two
@@ -299,11 +308,13 @@ typedef sem_t cksem_t;
 #define HAS_MD5		(control->flags & FLAG_MD5)
 #define CHECK_FILE	(control->flags & FLAG_CHECK)
 #define KEEP_BROKEN	(control->flags & FLAG_KEEP_BROKEN)
-#define LZO_TEST	(control->flags & FLAG_THRESHOLD)
+#define LZ4_TEST	(control->flags & FLAG_THRESHOLD)
 #define TMP_OUTBUF	(control->flags & FLAG_TMP_OUTBUF)
 #define TMP_INBUF	(control->flags & FLAG_TMP_INBUF)
 #define ENCRYPT		(control->flags & FLAG_ENCRYPT)
 
+#define IS_FROM_FILE ( !!(control->inFILE) && !STDIN )
+
 
 /* Structure to save state of computation between the single steps.  */
 struct md5_ctx
@@ -340,8 +351,21 @@ struct checksum {
 
 typedef i64 tag;
 
+struct node {
+	void *data;
+	struct node *prev;
+};
+
+struct runzip_node {
+	struct stream_info *sinfo;
+	pthread_t *pthreads;
+	struct runzip_node *prev;
+};
+
 struct rzip_state {
 	void *ss;
+	struct node *sslist;
+	struct node *head;
 	struct level *level;
 	tag hash_index[256];
 	struct hash_entry *hash_table;
@@ -401,6 +425,7 @@ struct rzip_control {
 	i64 max_mmap;
 	int threads;
 	char nice_val;		// added for consistency
+	int current_priority;
 	char major_version;
 	char minor_version;
 	i64 st_size;
@@ -416,6 +441,7 @@ struct rzip_control {
 	uchar *salt_pass;
 	int salt_pass_len;
 	uchar *hash;
+	char *passphrase;
 
 	pthread_mutex_t control_lock;
 	unsigned char eof;
@@ -431,10 +457,6 @@ struct rzip_control {
 	const char *util_infile;
 	char delete_infile;
 	const char *util_outfile;
-#define STREAM_BUCKET_SIZE 20
-	size_t sinfo_buckets;
-	size_t sinfo_idx;
-	struct stream_info **sinfo_queue;
 	char delete_outfile;
 	FILE *outputfile;
 	char library_mode;
@@ -450,6 +472,19 @@ struct rzip_control {
 	void (*next_tag)(rzip_control *, struct rzip_state *, i64, tag *);
 	tag (*full_tag)(rzip_control *, struct rzip_state *, i64);
 	i64 (*match_len)(rzip_control *, struct rzip_state *, i64, i64, i64, i64 *);
+
+	pthread_t *pthreads;
+	struct runzip_node *rulist;
+	struct runzip_node *ruhead;
+};
+
+struct uncomp_thread {
+	uchar *s_buf;
+	i64 u_len, c_len;
+	i64 last_head;
+	uchar c_type;
+	int busy;
+	int streamno;
 };
 
 struct stream {
@@ -475,6 +510,7 @@ struct stream_info {
 	i64 total_read;
 	i64 ram_alloced;
 	i64 size;
+	struct uncomp_thread *ucthreads;
 	long thread_no;
 	long next_thread;
 	int chunks;
